import java.util.ArrayList;
import java.util.List;

import gurobi.GRB;
import gurobi.GRBEnv;
import gurobi.GRBException;
import gurobi.GRBLinExpr;
import gurobi.GRBModel;
import gurobi.GRBVar;
/**
 * Solver class. This class contains only static methods. Used to solve instances using DRNDP method
 * @author Zhihao
 *
 */
class DRNDP {
	
	Network network;
	Instance instance;
	double L;			//hard lower bound on master problem
	
	GRBModel master;	//master problem
	GRBVar w;
	GRBVar[] x;
	GRBVar[] r;
	GRBVar[] v;
	
	GRBModel sub;		//sub problem
	GRBVar[] theta;
	GRBVar[] pi;
	GRBVar[][] rho;
	GRBVar[][] lambda;
	
	DRNDP(GRBEnv env,Network network_,Instance instance_,double L_) throws GRBException {
		network = network_;
		instance = instance_;
		L = L_;
		//MASTER PROBLEM
		master = new GRBModel(env);
		w = master.addVar(-GRB.INFINITY,GRB.INFINITY,0,GRB.CONTINUOUS,"z");
		x = SNDP.addXVars(master,network,instance);
		r = addRVars(master,network);
		v = addVVars(master,network);
		//construct the objective function
		GRBLinExpr masterobj = new GRBLinExpr();
		for(int ij = 0; ij < network.arcs.length; ij++) {
			masterobj.addTerm(network.arcs[ij].capCost,x[ij]);
		}
		masterobj.addTerm(1,w);
		for(int i = 0; i < network.demandNodes.length; i++) {
			masterobj.addTerm(instance.mu[i],r[i]);
			masterobj.addTerm(
					Math.pow(instance.mu[i],2)
					+instance.sigmasq[i]
					,v[i]);
		}
		master.addConstr(masterobj,GRB.GREATER_EQUAL,L,"lowerbound");	//set the hard lower bound for the objective function
		master.setObjective(masterobj,GRB.MINIMIZE);	//set the objective sense (also implicitly updates the model with the lower bound constraint)
		//SUBPROBLEM
		GRBEnv subEnv = new GRBEnv();
		subEnv.set(GRB.IntParam.OutputFlag,0);	//do not show output for sub problem
		//subEnv.set(GRB.IntParam.Presolve,0);	//do not presolve sub problem
		subEnv.set(GRB.IntParam.Method,0);	//use primal simplex for root node relaxation
		subEnv.set(GRB.IntParam.NodeMethod,0);	//use primal simplex for MIP node relaxations
		sub = new GRBModel(subEnv);
		theta = addThetaVars(sub,network);
		pi = addPiVars(sub,network);
		rho = addRhoVars(sub,network);
		lambda = addLambdaVars(sub,network);
		addFlowBalDualConstrs(sub,network,theta,pi);
		addMccormickConstrs(sub,network,pi,rho,lambda);
		addSos1Constrs(sub,network,rho);
	}
	/**
	 * Adds dual variables of the 1st moment matching constraints
	 */
	static GRBVar[] addRVars(GRBModel model,Network network) throws GRBException {
		GRBVar[] r = new GRBVar[network.demandNodes.length];
		for(int i = 0; i < network.demandNodes.length; i++) {
			r[i] = model.addVar(-GRB.INFINITY,GRB.INFINITY,0,GRB.CONTINUOUS,"r("+network.nodes[network.demandNodes[i]].name+")");
		}
		model.update();
		return r;
	}
	/**
	 * Adds dual variables of the 2nd moment matching constraints
	 */
	static GRBVar[] addVVars(GRBModel model,Network network) throws GRBException {
		GRBVar[] v = new GRBVar[network.demandNodes.length];
		for(int i = 0; i < network.demandNodes.length; i++) {
			v[i] = model.addVar(-GRB.INFINITY,GRB.INFINITY,0,GRB.CONTINUOUS,"v("+network.nodes[network.demandNodes[i]].name+")");
		}
		model.update();
		return v;
	}
	/**
	 * Adds dual variables of the capacity constraints
	 */
	static GRBVar[] addThetaVars(GRBModel model,Network network) throws GRBException {
		GRBVar[] theta = new GRBVar[network.arcs.length];
		for(int ij = 0; ij < network.arcs.length; ij++) {
			theta[ij] = model.addVar(0,GRB.INFINITY,0,GRB.CONTINUOUS,"theta("+network.arcs[ij].asString()+")");
		}
		return theta;
	}
	/**
	 * Adds dual variables of the flow balance constraints
	 */
	static GRBVar[] addPiVars(GRBModel model,Network network) throws GRBException {
		GRBVar[] pi = new GRBVar[network.nodes.length];
		for(int i = 0; i < network.supplyNodes.length; i++) {
			pi[network.supplyNodes[i]] = model.addVar(0,GRB.INFINITY,0,GRB.CONTINUOUS,"pi("+network.nodes[network.supplyNodes[i]].name+")");
		}
		for(int i = 0; i < network.transitNodes.length; i++) {
			pi[network.transitNodes[i]] = model.addVar(-GRB.INFINITY,GRB.INFINITY,0,GRB.CONTINUOUS,"pi("+network.nodes[network.transitNodes[i]].name+")");
		}
		for(int i = 0; i < network.demandNodes.length; i++) {
			pi[network.demandNodes[i]] = model.addVar(0,network.nodes[network.demandNodes[i]].penalty,0,GRB.CONTINUOUS,"pi("+network.nodes[network.demandNodes[i]].name+")");
		}
		model.update();
		return pi;
	}
	/**
	 * Adds sets of binary SOS1 variables to find the optimizer of the piecewise-linear approximation
	 */
	static GRBVar[][] addRhoVars(GRBModel model,Network network) throws GRBException {
		GRBVar[][] rho = new GRBVar[network.demandNodes.length][Main.numIntervals+1];
		for(int i = 0; i < network.demandNodes.length; i++) {
			for(int m = 0; m <= Main.numIntervals; m++) {
				rho[i][m] = model.addVar(0,1,0,GRB.BINARY,"rho("+network.nodes[network.demandNodes[i]].name+","+(m*Main.intervalLength)+")");
			}
		}
		model.update();
		return rho;
	}
	/**
	 * Adds the auxiliary binary variables for the McCormick inequalities
	 */
	static GRBVar[][] addLambdaVars(GRBModel model,Network network) throws GRBException {
		GRBVar[][] lambda = new GRBVar[network.demandNodes.length][Main.numIntervals+1];
		for(int i = 0; i < network.demandNodes.length; i++) {
			for(int m = 0; m <= Main.numIntervals; m++) {
				lambda[i][m] = model.addVar(0,GRB.INFINITY,0,GRB.CONTINUOUS,"lambda("+network.nodes[network.demandNodes[i]].name+","+(m*Main.intervalLength)+")");
			}
		}
		model.update();
		return lambda;
	}
	/**
	 * Adds the flow balance dual constraints
	 */
	static void addFlowBalDualConstrs(GRBModel model,Network network,GRBVar[] theta,GRBVar[] pi) throws GRBException {
		GRBLinExpr temp;
		for(int ij = 0; ij < network.arcs.length; ij++) {
			temp = new GRBLinExpr();
			temp.addTerm(1,pi[network.arcs[ij].destination]);
			temp.addTerm(-1,pi[network.arcs[ij].origin]);
			temp.addTerm(-1,theta[ij]);
			model.addConstr(temp,GRB.LESS_EQUAL,network.arcs[ij].flowCost,"flowbaldual("+network.arcs[ij].asString()+")");
		}
		model.update();
	}
	/**
	 * Adds the McCormick inequalities
	 */
	static void addMccormickConstrs(GRBModel model,Network network,GRBVar[] pi,GRBVar[][] rho,GRBVar[][] lambda) throws GRBException {
		GRBLinExpr temp;
		for(int i = 0; i < network.demandNodes.length; i++) {
			for(int m = 0; m <= Main.numIntervals; m++) {
				temp = new GRBLinExpr();
				temp.addTerm(1,lambda[i][m]);
				temp.addTerm(-1,pi[network.demandNodes[i]]);
				temp.addTerm(-network.nodes[network.demandNodes[i]].penalty,rho[i][m]);
				model.addConstr(temp,GRB.GREATER_EQUAL,-network.nodes[network.demandNodes[i]].penalty,"mccormickA("+network.nodes[network.demandNodes[i]].name+","+(m*Main.intervalLength)+")");
				temp = new GRBLinExpr();
				temp.addTerm(1,lambda[i][m]);
				temp.addTerm(-1,pi[network.demandNodes[i]]);
				model.addConstr(temp,GRB.LESS_EQUAL,0,"mccormickB("+network.nodes[network.demandNodes[i]].name+","+(m*Main.intervalLength)+")");
				temp = new GRBLinExpr();
				temp.addTerm(1,lambda[i][m]);
				temp.addTerm(-network.nodes[network.demandNodes[i]].penalty,rho[i][m]);
				model.addConstr(temp,GRB.LESS_EQUAL,0,"mccormickC("+network.nodes[network.demandNodes[i]].name+","+(m*Main.intervalLength)+")");
			}
		}
		model.update();
	}
	/**
	 * Adds the SOS1 constraints
	 */
	static void addSos1Constrs(GRBModel model,Network network,GRBVar[][] rho) throws GRBException {
		GRBLinExpr temp;
		for(int i = 0; i < network.demandNodes.length; i++) {
			temp = new GRBLinExpr();
			for(int m = 0; m <= Main.numIntervals; m++) {
				temp.addTerm(1,rho[i][m]);
			}
			model.addConstr(temp,GRB.EQUAL,1,"sos1("+network.nodes[network.demandNodes[i]].name+")");
		}
		model.update();
	}
	/**
	 * Update the (sub problem) objective function with values of the (master problem) variables x, r, and v
	 */
	static void updateSubObj(GRBModel model,GRBVar[] theta,GRBVar[] pi,GRBVar[][] rho,GRBVar[][] lambda,double[] xVal,double[] rVal,double[] vVal,Network network,Instance instance) throws GRBException {
		GRBLinExpr subobj = new GRBLinExpr();
		for(int i = 0; i < network.demandNodes.length; i++) {
			for(int m = 0; m <= Main.numIntervals; m++) {
				subobj.addTerm(m*Main.intervalLength,lambda[i][m]);
				subobj.addTerm(-m*Main.intervalLength*rVal[i]-Math.pow(m*Main.intervalLength,2)*vVal[i],rho[i][m]);
			}
		}
		for(int i = 0; i < network.supplyNodes.length; i++) {
			subobj.addTerm(-instance.supplies[i],pi[network.supplyNodes[i]]);
		}
		for(int ij = 0; ij < network.arcs.length; ij++) {
			double xvalue = 0;
			if(Math.abs(xVal[ij]) > 1e-6) {
				xvalue = xVal[ij];
			}
			subobj.addTerm(-xvalue,theta[ij]);
		}
		model.setObjective(subobj,GRB.MAXIMIZE);
		model.update();
	}
	/**
	 * Adds a cut (to the master problem) constructed from the (sub problem) variables theta, pi, rho, and lambda
	 */
	static void addCut(GRBModel model,GRBVar w,GRBVar[] x,GRBVar[] r,GRBVar[] v,double[] thetaVal,double[] piVal,double[][] rhoVal,double[][] lambdaVal,Network network,Instance instance,int cutIndex) throws GRBException {
		GRBLinExpr LHS = new GRBLinExpr();
		LHS.addTerm(1,w);
		for(int i = 0; i < network.demandNodes.length; i++) {
			for(int m = 0; m <= Main.numIntervals; m++) {
				LHS.addTerm(m*Main.intervalLength*rhoVal[i][m],r[i]);
				LHS.addTerm(Math.pow(m*Main.intervalLength,2)*rhoVal[i][m],v[i]);
			}
		}
		for(int ij = 0; ij < network.arcs.length; ij++) {
			LHS.addTerm(thetaVal[ij],x[ij]);
		}
		double RHS = 0;
		for(int i = 0; i < network.demandNodes.length; i++) {
			for(int m = 0; m <= Main.numIntervals; m++) {
				RHS += m*Main.intervalLength*lambdaVal[i][m];
			}
		}
		for(int i = 0; i < network.supplyNodes.length; i++) {
			RHS -= instance.supplies[i]*piVal[network.supplyNodes[i]];
		}
		model.addConstr(LHS,GRB.GREATER_EQUAL,RHS,"cut("+cutIndex+")");
		model.update();
	}
	
	/**
	 *  Solves an instance using the DRNDP approach
	 */
	Output solve() {
		List<Double> optVal = new ArrayList<Double>();
		List<double[]> xVal = new ArrayList<double[]>();
		List<Double> wVal = new ArrayList<Double>();
		List<double[]> rVal = new ArrayList<double[]>();
		List<double[]> vVal = new ArrayList<double[]>();
		List<double[]> thetaVal = new ArrayList<double[]>();
		List<double[]> piVal = new ArrayList<double[]>();
		List<double[][]> rhoVal = new ArrayList<double[][]>();
		List<double[][]> lambdaVal = new ArrayList<double[][]>();
		try {
			Output output = new Output();
			long startTime = System.currentTimeMillis();
			int numCuts = 0;
			boolean cont = true;
			while(cont) {
				cont = false;
				master.optimize();
				optVal.add(master.get(GRB.DoubleAttr.ObjVal));
				wVal.add(w.get(GRB.DoubleAttr.X));
				xVal.add(makeDoubleArray(x));
				rVal.add(makeDoubleArray(r));
				vVal.add(makeDoubleArray(v));
				updateSubObj(sub,theta,pi,rho,lambda,xVal.get(numCuts),rVal.get(numCuts),vVal.get(numCuts),network,instance);
				long hstart = System.currentTimeMillis();
				sub.optimize();
				output.htime += System.currentTimeMillis()-hstart;
				thetaVal.add(makeDoubleArray(theta));
				piVal.add(makeDoubleArray(pi));
				rhoVal.add(makeDoubleArray(rho));
				lambdaVal.add(makeDoubleArray(lambda));
				if(wVal.get(numCuts) < sub.get(GRB.DoubleAttr.ObjVal)-Math.abs(Main.drndpTol*sub.get(GRB.DoubleAttr.ObjVal))) {
					cont = true;
					addCut(master,w,x,r,v,thetaVal.get(numCuts),piVal.get(numCuts),rhoVal.get(numCuts),lambdaVal.get(numCuts),network,instance,numCuts);	//first cut is #0
					numCuts++;
				}
				if(numCuts%10 == 0) {
					System.out.println(numCuts+" cuts made");
					System.out.printf("     %.2f < %.2f\n",wVal.get(numCuts-1),sub.get(GRB.DoubleAttr.ObjVal)-Math.abs(Main.drndpTol*sub.get(GRB.DoubleAttr.ObjVal)));
				}
				/*if(numCuts > 490) {
					System.out.printf("			");
					for(int ij = 0; ij < network.arcs.length; ij++) {
						System.out.printf("%.2f	",xVal.get(numCuts-1)[ij]);
					}
					System.out.printf("\n");
					sub.write("C:/Users/Zhihao/Documents/Eclipse/workspace/DRNDP newest/src/data files and results/sub.lp");
					for(int ij = 0; ij < network.arcs.length; ij++) {
						System.out.printf("%.2f	",sub.getConstrByName("flowbaldual("+network.arcs[ij].asString()+")").get(GRB.DoubleAttr.Pi));
					}
					System.out.printf("\n");
				}*/
				if(sub.get(GRB.DoubleAttr.ObjVal) >1e15) {
					cont = false;
					master.write(Main.mainDirectory + "/master.lp");
					sub.write(Main.mainDirectory + "/sub.lp");
				}
			}
			//process solution
			output.solveTime = System.currentTimeMillis() - startTime;
			output.optVal = master.get(GRB.DoubleAttr.ObjVal);
			output.numCuts = numCuts;
			output.currentOpt = new double[optVal.size()];
			output.currentW = new double[wVal.size()];
			for(int m = 0; m < output.currentOpt.length; m++) {
				output.currentOpt[m] = optVal.get(m);
				output.currentW[m] = wVal.get(m);
			}
			output.currentX = xVal.toArray(new double[xVal.size()][]);
			output.currentR = rVal.toArray(new double[rVal.size()][]);
			output.currentV = vVal.toArray(new double[vVal.size()][]);
			output.currentTheta = thetaVal.toArray(new double[thetaVal.size()][]);
			output.currentPi = piVal.toArray(new double[piVal.size()][]);
			output.currentRho = rhoVal.toArray(new double[rhoVal.size()][][]);
			output.currentLambda = lambdaVal.toArray(new double[lambdaVal.size()][][]);
			output.currentCapCost = new double[output.currentX.length];
			for(int m = 0; m < output.currentX.length; m++) {
				for(int ij = 0; ij < network.arcs.length; ij++) {
					output.currentCapCost[m] += network.arcs[ij].capCost*output.currentX[m][ij];
				}
			}
			output.currentSecondStage = new double[output.currentR.length];
			for(int m = 0; m < output.currentR.length; m++) {
				output.currentSecondStage[m] += output.currentW[m];
				for(int i = 0; i < network.demandNodes.length; i++) {
					output.currentSecondStage[m] += (instance.mu[i]-instance.demandLBs[i])*output.currentR[m][i];
					output.currentSecondStage[m] += 
							(Math.pow(instance.mu[i],2)
							+instance.sigmasq[i]
							-Math.pow(instance.demandLBs[i],2))
							*output.currentV[m][i];
				}
			}
			output.x = output.currentX[numCuts];
			output.capCost = 0;
			for(int ij = 0; ij < network.arcs.length; ij++) {
				output.capCost += network.arcs[ij].capCost*output.x[ij];
			}
			output.secondStageSol = output.currentW[numCuts];
			for(int i = 0; i < network.demandNodes.length; i++) {
				output.secondStageSol += (instance.mu[i]-instance.demandLBs[i])*output.currentR[numCuts][i];
				output.secondStageSol += 
						(Math.pow(instance.mu[i],2)
						+instance.sigmasq[i]
						-Math.pow(instance.demandLBs[i],2))
						*output.currentV[numCuts][i];
			}
			return output;
		} catch (GRBException e) {
			System.err.println("Gurobi exception: " + e);
			return null;
		}
	}
	/**
	 * Outputs the solution of an array of variables into a double array
	 */
	private static double[] makeDoubleArray(GRBVar[] x) throws GRBException {
		double[] output = new double[x.length];
		for(int i = 0; i < x.length; i++) {
			if(x[i] != null) {
				output[i] = x[i].get(GRB.DoubleAttr.X);
			}
		}
		return output;
	}
	/**
	 * Outputs the solution of an array of variables into a double array
	 */
	private static double[][] makeDoubleArray(GRBVar[][] x) throws GRBException {
		double[][] output = new double[x.length][];
		for(int i = 0; i < x.length; i++) {
			output[i] = new double[x[i].length];
			for(int j = 0; j < x[i].length; j++) {
				if(x[i][j] != null) {
					output[i][j] = x[i][j].get(GRB.DoubleAttr.X);
				}
			}
		}
		return output;
	}
}
