import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintWriter;

/**
 * Data class. Instances are wrappers for multi-output functions
 * @author Zhihao, Hideaki
 *
 */
class Output {
	
	//common
	double[] x;						//capacity allocation
	long solveTime;					//solution time
	double optVal;					//optimal value
	double capCost;					//capacity allocation cost
	double secondStageSol;			//second stage objective value computed by model
	int numCuts = 0;				//number of cuts/iterations
	double[][] yQuant;				//flows in each scenario in reference set
	double[][] tQuant;				//unmet demand in each scenario in reference set
	double[] flowCostQuant;			//flow costs in each scenario in reference set
	double[] penaltyCostQuant;		//unmet demand penalty costs in each scenario in reference set
	//SNDP
	double[][] ySol;				//flows in each scenario in sample set
	double[][] tSol;				//unmet demand in each scenario in sample set
	double[] flowCostSol;			//flow costs in each scenario in sample set
	double[] penaltyCostSol;		//unmet demand penalty costs in each scenario in sample set
	//DRDNP
	double[] currentOpt;			//master problem optimal value at each iteration
	double[] currentCapCost;		//master problem capacity cost at each iteration
	double[] currentSecondStage;	//master problem second stage cost at each iteration
	double[][] currentX;			//x at each iteration
	double[] currentW;				//w at each iteration
	double[][] currentR;			//r at each iteration
	double[][] currentV;			//v at each iteration
	double[][] currentTheta;		//theta at each iteration
	double[][] currentPi;			//pi at each iteration
	double[][][] currentRho;		//rho at each iteration
	double[][][] currentLambda;		//lambda at each iteration
	
	//RNDP added by Hideaki
	double[][] currentQ;				//q at each iteration
	double[][] currentU;				//u at each iteration
	long htime;							//time spent for solving subproblem
	
	Output() {}
	
	/**
	 * Prints outputs for the SNDP to a folder
	 */
	void printSNDP(String folderName) {
		try {
			printSummary(folderName+"/summary.csv");
			print2d(ySol,folderName+"/ySol.csv");
			print2d(tSol,folderName+"/tSol.csv");
			print1d(flowCostSol,folderName+"/flowCostSol.csv");
			print1d(penaltyCostSol,folderName+"/penaltyCostSol.csv");
		} catch (IOException e) {
			System.err.println("I/O exception: " + e);
		}
	}
	/**
	 * Prints outputs for the DRNDP to a folder
	 */
	void printDRNDP(String folderName) {
		try {
			printSummary(folderName+"/summary.csv");
			print1d(currentOpt,folderName+"/currentOpt.csv");
			print1d(currentCapCost,folderName+"/currentCapCost.csv");
			print1d(currentSecondStage,folderName+"/currentSecondStage.csv");
			print2d(currentX,folderName+"/currentX.csv");
			print1d(currentW,folderName+"/currentW.csv");
			print2d(currentR,folderName+"/currentR.csv");
			print2d(currentV,folderName+"/currentV.csv");
			print2d(currentTheta,folderName+"/currentTheta.csv");
			print2d(currentPi,folderName+"/currentPi.csv");
			print3d(currentRho,folderName+"/currentRho.csv");
			print3d(currentLambda,folderName+"/currentLambda.csv");
		} catch (IOException e) {
			System.err.println("I/O exception: " + e);
		}
	}
	/** added by Hideaki
	 * Prints outputs for the RNDP to a folder
	 */
	void printRNDP(String folderName) {
		try {
			printSummary(folderName+"/summary.csv");
			print1d(currentOpt,folderName+"/currentOpt.csv");
			print1d(currentCapCost,folderName+"/currentCapCost.csv");
			print1d(currentSecondStage,folderName+"/currentSecondStage.csv");
			print2d(currentX,folderName+"/currentX.csv");
			print1d(currentW,folderName+"/currentW.csv");
			print2d(currentTheta,folderName+"/currentTheta.csv");
			print2d(currentPi,folderName+"/currentPi.csv");
			print2d(currentQ,folderName+"/currentQ.csv");
			print2d(currentU,folderName+"/currentU.csv");
		} catch (IOException e) {
			System.err.println("I/O exception: " + e);
		}
	}
	/**
	 * Prints summary statistics to a file
	 */
	void printSummary(String filename) throws IOException {
		PrintWriter writer = new PrintWriter(filename);
		for(int i = 0; i < x.length; i++) {
			writer.printf(x[i]+",");
		}
		writer.printf("\n");
		writer.println(solveTime);
		writer.println(optVal);
		writer.println(capCost);
		writer.println(secondStageSol);
		writer.println(numCuts);
		writer.println(htime);
		writer.close();
	}
	/**
	 * Prints performance statistics to a folder
	 */
	void printPerformance(String folderName) {
		try {
			PrintWriter writer = new PrintWriter(folderName+"/performance summary.csv");
			for(int i = 0; i < x.length; i++) {
				writer.printf(x[i]+",");
			}
			writer.printf("\n");
			double meanFlowCost = 0;
			for(int s = 0; s < flowCostQuant.length; s++) {
				meanFlowCost += flowCostQuant[s];
			}
			meanFlowCost /= flowCostQuant.length;
			double meanPenaltyCost = 0;
			for(int s = 0; s < penaltyCostQuant.length; s++) {
				meanPenaltyCost += penaltyCostQuant[s];
			}
			meanPenaltyCost /= penaltyCostQuant.length;
			writer.println(capCost+meanFlowCost+meanPenaltyCost);
			writer.println(capCost);
			writer.println(meanFlowCost);
			writer.println(meanPenaltyCost);
			writer.close();
			print2d(yQuant,folderName+"/yQuant.csv");
			print2d(tQuant,folderName+"/tQuant.csv");
			print1d(flowCostQuant,folderName+"/flowCostQuant.csv");
			print1d(penaltyCostQuant,folderName+"/penaltyCostQuant.csv");
		} catch (IOException e) {
			System.err.println("I/O exception: " + e);
		}
	}
	/**
	 * Prints a 3D array to a file
	 */
	void print3d(double[][][] array,String filename) throws IOException {
		PrintWriter writer = new PrintWriter(filename);
		for(int s = 0; s < array.length; s++) {
			for(int i = 0; i < array[s].length; i++) {
				for(int m = 0; m < array[s][i].length; m++) {
					writer.printf(array[s][i][m]+",");
				}
				writer.printf("\n");
			}
			writer.printf("\n");
		}
		writer.close();
	}
	/**
	 * Prints a 2D array to a file
	 */
	void print2d(double[][] array,String filename) throws IOException {
		PrintWriter writer = new PrintWriter(filename);
		for(int s = 0; s < array.length; s++) {
			for(int i = 0; i < array[s].length; i++) {
				writer.printf(array[s][i]+",");
			}
			writer.printf("\n");
		}
		writer.close();
	}
	/**
	 * Prints a 1D array to a file
	 */
	void print1d(double[] print1d,String filename) throws IOException {
		PrintWriter writer = new PrintWriter(filename);
		for(int s = 0; s < print1d.length; s++) {
			writer.println(print1d[s]);
		}
		writer.close();
	}
	/**
	 * Reads a 1D array from a file
	 */
	double[] readXfromSummary(String filename) {
		try {
			BufferedReader br = new BufferedReader(new FileReader(filename));
			String[] splitstr = br.readLine().split(",");
			double[] x = new double[splitstr.length];
			for(int i = 0; i < x.length; i++) {
				x[i] = Double.parseDouble(splitstr[i]);
			}
			br.close();
			return x;
		} catch (IOException e) {
			System.err.println("I/O exception: " + e);
			return null;
		}
	}
}
