import java.util.ArrayList;
import java.util.List;

import gurobi.GRB;
import gurobi.GRBEnv;
import gurobi.GRBException;
import gurobi.GRBLinExpr;
import gurobi.GRBModel;
import gurobi.GRBVar;
/**
 * Solver class. This class contains only static methods. Used to solve instances using RNDP method
 * @author Hideaki
 *
 */
class RNDP {
	
	Network network;
	Instance instance;
	static int Gamma;	//threshold of cardinality  //Must figure out how to pass value of Gamma
	
	GRBModel master;	//master problem
	GRBVar w;
	GRBVar[] x;
	
	GRBModel sub;		//sub problem
	GRBVar[] theta;
	GRBVar[] pi;
	GRBVar[] q;
	GRBVar[] u;

	
	RNDP(GRBEnv env,Network network_,Instance instance_,double cardinality) throws GRBException {
		network = network_;
		instance = instance_;
		Gamma = (int)(cardinality * instance.mu.length);
		//MASTER PROBLEM
		master = new GRBModel(env);
		w = master.addVar(0,GRB.INFINITY,0,GRB.CONTINUOUS,"z");
		x = SNDP.addXVars(master,network,instance);
		master.update();
		//construct the objective function
		GRBLinExpr masterobj = new GRBLinExpr();
		for(int ij = 0; ij < network.arcs.length; ij++) {
			masterobj.addTerm(network.arcs[ij].capCost,x[ij]);
		}
		masterobj.addTerm(1,w);
		master.setObjective(masterobj,GRB.MINIMIZE);	//set the objective sense (also implicitly updates the model with the lower bound constraint)
		//SUBPROBLEM
		GRBEnv subEnv = new GRBEnv();
		subEnv.set(GRB.IntParam.OutputFlag,0);	//do not show output for sub problem
		//subEnv.set(GRB.IntParam.Presolve,0);	//do not presolve sub problem
		subEnv.set(GRB.IntParam.Method,0);	//use primal simplex for root node relaxation
		subEnv.set(GRB.IntParam.NodeMethod,0);	//use primal simplex for MIP node relaxations
		sub = new GRBModel(subEnv);
		theta = DRNDP.addThetaVars(sub,network);
		pi = DRNDP.addPiVars(sub,network);
		q = addQVars(sub,network);
		u = addUVars(sub,network);
		addFlowBalDualConstrs(sub,network,theta,pi);
		addMccormickConstrsRobust(sub,network,pi,q,u);
		addCardinalityConstrs(sub,network,q);
	}
	
	/**
	 * Adds binary variables indicating the fluctuation of variables
	 */
	static GRBVar[] addQVars(GRBModel model,Network network) throws GRBException {
		GRBVar[] q = new GRBVar[network.demandNodes.length];
		for(int i = 0; i < network.demandNodes.length; i++) {
			q[i] = model.addVar(0,1,0,GRB.BINARY,"q("+network.nodes[network.demandNodes[i]].name+")");
		}
		model.update();
		return q;
	}
	/**
	 * Adds the auxiliary binary variables for the McCormick inequalities of pi and q
	 */
	static GRBVar[] addUVars(GRBModel model,Network network) throws GRBException {
		GRBVar[] u = new GRBVar[network.demandNodes.length];
		for(int i = 0; i < network.demandNodes.length; i++) {
			u[i] = model.addVar(0,GRB.INFINITY,0,GRB.CONTINUOUS,"u("+network.nodes[network.demandNodes[i]].name+")");
		}
		model.update();
		return u;
	}
	/**
	 * Adds the flow balance dual constraints
	 */
	static void addFlowBalDualConstrs(GRBModel model,Network network,GRBVar[] theta,GRBVar[] pi) throws GRBException {
		GRBLinExpr temp;
		for(int ij = 0; ij < network.arcs.length; ij++) {
			temp = new GRBLinExpr();
			temp.addTerm(1,pi[network.arcs[ij].destination]);
			temp.addTerm(-1,pi[network.arcs[ij].origin]);
			temp.addTerm(-1,theta[ij]);
			model.addConstr(temp,GRB.LESS_EQUAL,network.arcs[ij].flowCost,"flowbaldual("+network.arcs[ij].asString()+")");
		}
		model.update();
	}
	/**
	 * Adds the McCormick inequalities
	 */
	static void addMccormickConstrsRobust(GRBModel model,Network network,GRBVar[] pi,GRBVar[] q,GRBVar[] u) throws GRBException {
		GRBLinExpr temp;
		for(int i = 0; i < network.demandNodes.length; i++) {
			temp = new GRBLinExpr();
			temp.addTerm(1,u[i]);
			temp.addTerm(-1,pi[network.demandNodes[i]]);
			temp.addTerm(-network.nodes[network.demandNodes[i]].penalty,q[i]);
			model.addConstr(temp,GRB.GREATER_EQUAL,-network.nodes[network.demandNodes[i]].penalty,"mccormickRobustA("+network.nodes[network.demandNodes[i]].name+")");
			temp = new GRBLinExpr();
			temp.addTerm(1,u[i]);
			temp.addTerm(-1,pi[network.demandNodes[i]]);
			model.addConstr(temp,GRB.LESS_EQUAL,0,"mccormickRobustB("+network.nodes[network.demandNodes[i]].name+")");
			temp = new GRBLinExpr();
			temp.addTerm(1,u[i]);
			temp.addTerm(-network.nodes[network.demandNodes[i]].penalty,q[i]);
			model.addConstr(temp,GRB.LESS_EQUAL,0,"mccormickRobustC("+network.nodes[network.demandNodes[i]].name+")");
		}
		model.update();
	}
	/**
	 * Adds the Cardinality constraint
	 */
	static void addCardinalityConstrs(GRBModel model,Network network,GRBVar[]q) throws GRBException {
		GRBLinExpr temp = new GRBLinExpr();
		for(int i = 0; i < network.demandNodes.length; i++) {
			temp.addTerm(1,q[i]);
		}
		model.addConstr(temp,GRB.LESS_EQUAL,Gamma,"Cardinality");
		model.update();
	}
	/**
	 * Update the (sub problem) objective function with values of the (master problem) variables x, r, and v
	 */
	static void updateSubObjRobust(GRBModel model,GRBVar[] theta,GRBVar[] pi,GRBVar[] q,GRBVar[] u, double[] xVal, Network network,Instance instance) throws GRBException {
		GRBLinExpr subobj = new GRBLinExpr();
		for(int i = 0; i < network.demandNodes.length; i++) {
			//subobj.addTerm((instance.demandUBs[i]+instance.demandLBs[i])/2,pi[network.demandNodes[i]]);
			subobj.addTerm(instance.mu[i],pi[network.demandNodes[i]]);
			//subobj.addTerm((instance.demandUBs[i]-instance.demandLBs[i])/2,u[i]);
			subobj.addTerm(instance.demandUBs[i]-instance.mu[i],u[i]);
		}
		for(int i = 0; i < network.supplyNodes.length; i++) {
			subobj.addTerm(-instance.supplies[i],pi[network.supplyNodes[i]]);
		}
		for(int ij = 0; ij < network.arcs.length; ij++) {
			double xvalue = 0;
			if(Math.abs(xVal[ij]) > 1e-6) {
				xvalue = xVal[ij];
			}
			subobj.addTerm(-xvalue,theta[ij]);
		}
		model.setObjective(subobj,GRB.MAXIMIZE);
		model.update();
	}
	/**
	 * Adds a cut (to the master problem) constructed from the (sub problem) variables theta, pi, rho, and lambda
	 */
	static void addCutRobust(GRBModel model,GRBVar w,GRBVar[] x, double[] thetaVal,double[] piVal,double[] uVal,Network network,Instance instance,int cutIndex) throws GRBException {
		GRBLinExpr LHS = new GRBLinExpr();
		LHS.addTerm(1,w);
		for(int ij = 0; ij < network.arcs.length; ij++) {
			LHS.addTerm(thetaVal[ij],x[ij]);
		}
		double RHS = 0;
		for(int i = 0; i < network.demandNodes.length; i++) {
			//RHS += (instance.demandUBs[i]+instance.demandLBs[i])/2*piVal[network.demandNodes[i]] + (instance.demandUBs[i]-instance.demandLBs[i])/2*uVal[i]; 
			RHS += instance.mu[i]*piVal[network.demandNodes[i]] + (instance.demandUBs[i]-instance.mu[i])*uVal[i]; 
		}
		for(int i = 0; i < network.supplyNodes.length; i++) {
			RHS -= instance.supplies[i]*piVal[network.supplyNodes[i]];
		}
		model.addConstr(LHS,GRB.GREATER_EQUAL,RHS,"cutRobust("+cutIndex+")");
		model.update();
	}
	
	/**
	 *  Solves an instance using the DRNDP approach
	 */
	Output solve() {
		List<Double> optVal = new ArrayList<Double>();
		List<double[]> xVal = new ArrayList<double[]>();
		List<Double> wVal = new ArrayList<Double>();
		List<double[]> thetaVal = new ArrayList<double[]>();
		List<double[]> piVal = new ArrayList<double[]>();
		List<double[]> qVal = new ArrayList<double[]>();
		List<double[]> uVal = new ArrayList<double[]>();
		try {
			Output output = new Output();
			long startTime = System.currentTimeMillis();
			int numCuts = 0;
			boolean cont = true;
			while(cont) {
				cont = false;
				master.optimize();
				optVal.add(master.get(GRB.DoubleAttr.ObjVal));
				wVal.add(w.get(GRB.DoubleAttr.X));
				xVal.add(makeDoubleArray(x));
				updateSubObjRobust(sub,theta,pi,q,u,xVal.get(numCuts),network,instance);
				long hstart = System.currentTimeMillis();
				sub.optimize();
				output.htime += System.currentTimeMillis()-hstart;
				thetaVal.add(makeDoubleArray(theta));
				piVal.add(makeDoubleArray(pi));
				qVal.add(makeDoubleArray(q));
				uVal.add(makeDoubleArray(u));
				if(wVal.get(numCuts) < sub.get(GRB.DoubleAttr.ObjVal)-Math.abs(Main.drndpTol*sub.get(GRB.DoubleAttr.ObjVal))) {
					cont = true;
					addCutRobust(master,w,x,thetaVal.get(numCuts),piVal.get(numCuts),uVal.get(numCuts),network,instance,numCuts);	//first cut is #0
					numCuts++;
				}
				if(numCuts%10 == 0) {
					System.out.println(numCuts+" cuts made");
					System.out.printf("     %.2f < %.2f\n",wVal.get(numCuts-1),sub.get(GRB.DoubleAttr.ObjVal)-Math.abs(Main.drndpTol*sub.get(GRB.DoubleAttr.ObjVal)));
				}
				//in case it doesn't stop
				if(numCuts == 3000) {
					cont = false;
					master.write(Main.mainDirectory + "/master.lp");
					sub.write(Main.mainDirectory + "/sub.lp");
				}
			}
			//process solution
			output.solveTime = System.currentTimeMillis() - startTime;
			output.optVal = master.get(GRB.DoubleAttr.ObjVal);
			output.numCuts = numCuts;
			output.currentOpt = new double[optVal.size()];
			output.currentW = new double[wVal.size()];
			for(int m = 0; m < output.currentOpt.length; m++) {
				output.currentOpt[m] = optVal.get(m);
				output.currentW[m] = wVal.get(m);
			}
			output.currentX = xVal.toArray(new double[xVal.size()][]);
			output.currentTheta = thetaVal.toArray(new double[thetaVal.size()][]);
			output.currentPi = piVal.toArray(new double[piVal.size()][]);
			output.currentQ = qVal.toArray(new double[qVal.size()][]);
			output.currentU = uVal.toArray(new double[uVal.size()][]);
			output.currentCapCost = new double[output.currentX.length];
			for(int m = 0; m < output.currentX.length; m++) {
				for(int ij = 0; ij < network.arcs.length; ij++) {
					output.currentCapCost[m] += network.arcs[ij].capCost*output.currentX[m][ij];
				}
			}
			output.currentSecondStage = new double[output.currentW.length];
			for(int m = 0; m < output.currentW.length; m++) {
				output.currentSecondStage[m] += output.currentW[m];
			}
			output.x = output.currentX[numCuts];
			output.capCost = 0;
			for(int ij = 0; ij < network.arcs.length; ij++) {
				output.capCost += network.arcs[ij].capCost*output.x[ij];
			}
			output.secondStageSol = output.currentW[numCuts];
			return output;
		} catch (GRBException e) {
			System.err.println("Gurobi exception: " + e);
			return null;
		}
	}
	/**
	 * Outputs the solution of an array of variables into a double array
	 */
	private static double[] makeDoubleArray(GRBVar[] x) throws GRBException {
		double[] output = new double[x.length];
		for(int i = 0; i < x.length; i++) {
			if(x[i] != null) {
				output[i] = x[i].get(GRB.DoubleAttr.X);
			}
		}
		return output;
	}
}
